#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from akangatu.distep import DIStep


class Slice(DIStep):
    """Select rows(or columns) inside a given interval, including limits."""

    def __init__(self, code=":5", fields="X,Y"):
        # REMINDER: super call with locals leads to infinite loop
        super().__init__(code=code, fields=fields)

        # https://stackoverflow.com/a/51105983/9681577
        def toslice(txt):
            return slice(*map(lambda x: int(x.strip()) if x.strip() else None, txt.split(':')))

        self.slice = tuple(toslice(txt) for txt in code.split(","))
        self.fields = fields.split(",")

    def _process_(self, data):
        newfields = {k: (lambda k_: lambda: data[k_][self.slice])(k) for k in self.fields}
        return data.update(self, **newfields)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.
#

from unittest import TestCase

from pandas import DataFrame

from aiuna.content.root import Root
from aiuna.step.dataset import Dataset

from kururu.tool.dataflow.autoins import AutoIns
from kururu.tool.enhancement.instance.sampling.over.rnd import ROS_, ROS
from kururu.tool.manipulation.slice import Slice


class Test(TestCase):
    def test__ros_(self):
        iris = Root >> Dataset()
        truncated_iris = iris >> Slice(last=119)
        rebalanced_iris = truncated_iris >> ROS_
        self.assertEqual(50, DataFrame(rebalanced_iris.y).value_counts()["virginica"])
        larger_iris = iris >> ROS_(strategy={"virginica": 100, "versicolor": 50, "setosa": 50})
        self.assertEqual(200, len(larger_iris.X))

    def test__ros(self):
        iris = Root >> (Dataset() * AutoIns)
        truncated_iris = iris >> Slice(last=119)
        rebalanced_iris = truncated_iris >> ROS
        self.assertEqual(50, DataFrame(rebalanced_iris.inner.y).value_counts()["virginica"])
        larger_iris = iris >> ROS(strategy={"virginica": 100, "versicolor": 50, "setosa": 50})
        self.assertEqual(200, len(larger_iris.inner.X))



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from akangatu.operator.unary.inop import In
from kururu.tool.dataflow.autoins import AutoIns
from kururu.tool.evaluation.mixin.partitioning import withPartitioning
from kururu.tool.evaluation.split import Split1
from akangatu.distep import DIStep
from akangatu.abs.mixin.fixedparam import asFixedParam


class Partition(asFixedParam, withPartitioning, DIStep):
    def __init__(self, mode="cv", splits=10, test_size=0.3, seed=0, fields="X,Y"):
        config = locals().copy()
        del config["self"]
        withPartitioning.__init__(self, mode, config)
        DIStep.__init__(self, **config)

    def _process_(self, data):
        """"""
        idxs_func = lambda: self.partitionings(data)

        # REMINDER: Partition faz data virar data com stream; e nesse stream vem data transformado e com inner.
        def gen():
            idxs = idxs_func()
            for i in range(self.splits):
                # print("partition", i, end="   ")
                trainS = Split1(i, "train", **self.config, _indices=idxs[i][0])
                testS = Split1(i, "test", **self.config, _indices=idxs[i][1])
                wk = AutoIns * In(trainS) * testS
                yield wk.process(data)

        return data.update(self, stream=gen)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from scipy.stats import rankdata
from akangatu.distep import DIStep


class Rank(DIStep):
    """Rank all columns(or rows) independently, or by a given column/row that must exist in all matrices."""

    def __init__(self, fields="U", byindex=0, along="rows"):
        super().__init__(fields=fields, byindex=byindex, along=along)
        self.fields = fields.split(",")
        self.along = along
        self.byindex = byindex
        if along not in ["rows", "cols"]:
            print(self.name, f"Unknown type of entry to rank: by='{along}'. Should be 'rows' or 'cols'.")
            exit()

    def _process_(self, data):
        dic = {}
        for field in self.fields:
            def result():
                M = data[field]
                if self.byindex is not None:
                    M = M[:, self.byindex] if self.along == "rows" else M[self.byindex, :]
                if self.along == "rows":
                    return (rankdata(M, axis=0) - 1).astype(int)
                return (rankdata(M, axis=1) - 1).astype(int)

            dic["r" + field] = result
        return data.update(self, **dic)



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

from sklearn.decomposition import PCA as PCAsk

from aiuna.step.delete import Del
from akangatu.abs.mixin.macro import asMacro
from akangatu.operator.unary.inop import In
from kururu.tool.dataflow.autoins import AutoIns
from kururu.tool.enhancement.attribute.abs.transformer import Transformer
from akangatu.transf.config import globalcache


class PCAo(Transformer):
    """Reduce dimensionality by PCA. Use inner data as "training" data, but transform only the outer data.

    Hint: use PCA to transform both inner and outer data."""

    def __init__(self, inner=None, n=2, seed=0):
        self.n = n
        self.seed = seed
        super().__init__(inner, n=n, seed=seed)

    def translate(self, exception, data):
        msg = str(exception)
        if "n_components=" in msg and "must be between 0 and min(n_samples, n_features)=" in msg:
            return f"n:{self.n}"  # > Xw({data.X.shape[1]}) or n:{self.n} > Xh({data.X.shape[0]})"

    def _algorithm_func(self):
        return lambda: PCAsk(n_components=self.n, random_state=self.seed)


class PCAb(asMacro, PCAo):
    """Apply PCA to transform both inner and outer data. Inner data is used as "training" data."""

    def _step_(self):
        pca = PCAo(**self.held)
        return pca * In(AutoIns * pca * Del("inner"))


pca = PCAb()

# l = []
# for i in range(1, 28):
#     l.append(round(pow(1.2, i)))
# print(sorted(list(set(l))))



#  Copyright (c) 2020. Davi Pereira dos Santos
#  This file is part of the kururu project.
#  Please respect the license - more about this in the section (*) below.
#
#  kururu is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  kururu is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with kururu.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is a crime and is unethical regarding the effort and
#  time spent here.
#  Relevant employers or funding agencies will be notified accordingly.

import setuptools

NAME = "kururu"

VERSION = "0.2102.21"

AUTHOR = 'Davi Pereira-Santos'

AUTHOR_EMAIL = 'dpsabc@gmail.com'

DESCRIPTION = 'Data Science'

with open('README.md', 'r') as fh:
    LONG_DESCRIPTION = fh.read()

LICENSE = 'GPL3'

URL = 'https://github.com/davips/kururu'

DOWNLOAD_URL = 'https://github.com/davips/kururu/releases'

CLASSIFIERS = ['Intended Audience :: Science/Research',
               'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
               'Natural Language :: English',
               'Programming Language :: Python',
               'Topic :: Scientific/Engineering',
#posix               'Operating System :: Linux',
               'Programming Language :: Python :: 3.8']

INSTALL_REQUIRES = [
    'numpy', 'sklearn', 'liac-arff', 'imblearn', 'tatu'
]

EXTRAS_REQUIRE = {
}

SETUP_REQUIRES = ['wheel']

setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    description=DESCRIPTION,
    download_url=DOWNLOAD_URL,
    extras_require=EXTRAS_REQUIRE,
    install_requires=INSTALL_REQUIRES,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    packages=setuptools.find_packages(),
    setup_requires=SETUP_REQUIRES,
    url=URL,
    # scripts=['xxxxx']
)

package_dir = {'': 'kururu'}  # So IDEs (e.g. Intellij) can recognize the working dir when running with ctr+shift+F10.
